
// Copyright 2019 Shlomi Steinberg
// 

#pragma once

#include <array>
#include "glm.hpp"


class illuminant_base {
protected:
	illuminant_base() = default;

public:
	// Wavelength in nm
	virtual float operator()(float wavelength) const noexcept = 0;
	virtual float operator()(float min_wavelength, float max_wavelength) const noexcept = 0;
	virtual float draw(const float seed) const noexcept = 0;
	virtual glm::vec3 XYZtoRGB(const glm::vec3 &XYZ) const noexcept = 0;
    virtual float coherence_length() const noexcept = 0;
};

template <typename Data>
class illuminant : public illuminant_base {
private:
	std::array<float, 100> icdf;

public:
	illuminant() noexcept {
		std::array<float, Data::data.size()> cdf;
		cdf[0] = Data::data[0] * Data::normalizer;
		for (std::size_t idx=1; idx<Data::data.size(); ++idx)
			cdf[idx] = cdf[idx-1] + Data::data[idx] * Data::normalizer;

		std::size_t p=0;
		for (std::size_t idx=0; idx<100; ++idx) {
			while (p+1<cdf.size() && cdf[p+1]<=idx*.01f) ++p;
			icdf[idx] = p; 
			icdf[idx] = glm::mix(Data::min_wl, Data::max_wl, icdf[idx]/(cdf.size()-1));
		}
	}

	// Wavelength in nm
	float operator()(float wavelength) const noexcept override {
		const auto t = glm::max(0.f, (wavelength - Data::min_wl) / (Data::max_wl-Data::min_wl) * Data::data.size());
		const auto l = glm::min(Data::data.size()-1, static_cast<std::size_t>(t));
		const auto h = glm::min(Data::data.size()-1, static_cast<std::size_t>(t)+1);

		return glm::mix(Data::data[l],Data::data[h],glm::fract(t)) * Data::normalizer;
	}
	float operator()(float min_wavelength, float max_wavelength) const noexcept override {
		const auto t0 = glm::max(0.f, (min_wavelength - Data::min_wl) / (Data::max_wl-Data::min_wl) * Data::data.size());
		const auto t1 = glm::max(0.f, (max_wavelength - Data::min_wl) / (Data::max_wl-Data::min_wl) * Data::data.size());
		const auto l0 = glm::min(Data::data.size()-1, static_cast<std::size_t>(t0));
		const auto l  = glm::min(Data::data.size()-1, static_cast<std::size_t>(t0)+1);
		const auto h  = glm::min(Data::data.size()-1, static_cast<std::size_t>(t1));
		const auto h1 = glm::min(Data::data.size()-1, static_cast<std::size_t>(t1)+1);

		auto power = .0f;
		if (h>l) {
			for (auto idx=l;idx<h;++idx)
				power += glm::mix(Data::data[idx],Data::data[idx+1],.5f);
			power/=static_cast<float>(h-l);
		}
		power += glm::mix(Data::data[l0], Data::data[l], glm::fract(glm::mix(static_cast<float>(l0),static_cast<float>(l), .5f))) * (1.f - glm::fract(t0));
		power += glm::mix(Data::data[h], Data::data[h1], glm::fract(glm::mix(static_cast<float>(h) ,static_cast<float>(h1),.5f))) * glm::fract(t1);

		return power * Data::normalizer;
	}
	float draw(const float seed) const noexcept override {
		const auto idx = glm::clamp(.0f,icdf.size()-.5f,seed*icdf.size());
		const auto l = icdf[static_cast<std::size_t>(idx)];
		const auto h = icdf[glm::min(static_cast<std::size_t>(idx)+1,icdf.size())];
		return glm::mix(l,h,glm::fract(idx));
	}

	glm::vec3 XYZtoRGB(const glm::vec3 &XYZ) const noexcept override { return XYZ * Data::template XYZtoRGB<float>; }
    float coherence_length() const noexcept override { return Data::coherence_length; }
};


struct illuminant_d65_data {
	static constexpr std::size_t min_wl = 300;
	static constexpr std::size_t max_wl = 780;
	static constexpr float normalizer = 1.f / 7561.f;
	static constexpr auto data = std::array<float, 97>{
		0.f,
		1.f,
		3.f,
		11.f,
		20.f,
		28.f,
		37.f,
		38.f,
		39.f,
		42.f,
		44.f,
		45.f,
		46.f,
		49.f,
		52.f,
		51.f,
		49.f,
		52.f,
		54.f,
		68.f,
		82.f,
		87.f,
		91.f,
		92.f,
		93.f,
		90.f,
		86.f,
		95.f,
		104.f,
		110.f,
		117.f,
		117.f,
		117.f,
		116.f,
		114.f,
		115.f,
		115.f,
		112.f,
		108.f,
		109.f,
		109.f,
		108.f,
		107.f,
		106.f,
		104.f,
		106.f,
		107.f,
		106.f,
		104.f,
		104.f,
		104.f,
		102.f,
		100.f,
		98.f,
		96.f,
		96.f,
		95.f,
		92.f,
		88.f,
		89.f,
		90.f,
		89.f,
		89.f,
		88.f,
		87.f,
		85.f,
		83.f,
		83.f,
		83.f,
		81.f,
		80.f,
		80.f,
		80.f,
		81.f,
		82.f,
		80.f,
		78.f,
		74.f,
		69.f,
		70.f,
		71.f,
		72.f,
		74.f,
		67.f,
		61.f,
		65.f,
		69.f,
		72.f,
		75.f,
		69.f,
		63.f,
		55.f,
		46.f,
		56.f,
		66.f,
		65.f,
		63.f,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_d65 = illuminant<illuminant_d65_data>;


struct illuminant_d50_data {
	static constexpr std::size_t min_wl = 300;
	static constexpr std::size_t max_wl = 780;
	static constexpr float normalizer = 1.f / 7.2007e+03f;
	static constexpr auto data = std::array<float, 97>{
		0.019f,
		1.035f,
		2.051f,
		4.914f,
		7.778f,
		11.263f,
		14.748f,
		16.348f,
		17.948f,
		19.479f,
		21.010f,
		22.476f,
		23.942f,
		25.451f,
		26.961f,
		25.724f,
		24.488f,
		27.179f,
		29.871f,
		39.589f,
		49.308f,
		52.910f,
		56.513f,
		58.273f,
		60.034f,
		58.926f,
		57.818f,
		66.321f,
		74.825f,
		81.036f,
		87.247f,
		88.930f,
		90.612f,
		90.990f,
		91.368f,
		93.238f,
		95.109f,
		93.536f,
		91.963f,
		93.843f,
		95.724f,
		96.169f,
		96.613f,
		96.871f,
		97.129f,
		99.614f,
		102.099f,
		101.427f,
		100.755f,
		101.536f,
		102.317f,
		101.159f,
		100.000f,
		98.868f,
		97.735f,
		98.327f,
		98.918f,
		96.208f,
		93.499f,
		95.593f,
		97.688f,
		98.478f,
		99.269f,
		99.155f,
		99.042f,
		97.382f,
		95.722f,
		97.290f,
		98.857f,
		97.262f,
		95.667f,
		96.929f,
		98.190f,
		100.597f,
		103.003f,
		101.068f,
		99.133f,
		93.257f,
		87.381f,
		89.492f,
		91.604f,
		92.246f,
		92.889f,
		84.872f,
		76.854f,
		81.683f,
		86.511f,
		89.546f,
		92.580f,
		85.405f,
		78.230f,
		67.961f,
		57.692f,
		70.307f,
		82.923f,
		80.599f,
		78.274f,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_d50 = illuminant<illuminant_d50_data>;


struct illuminant_d75_data {
	static constexpr std::size_t min_wl = 300;
	static constexpr std::size_t max_wl = 800;
	static constexpr float normalizer = 1.f / 8017.76f;
	static constexpr auto data = std::array<float, 101>{
        0.043,
        2.588,
        5.133,
        17.470,
        29.808,
        42.369,
        54.930,
        56.095,
        57.259,
        60.000,
        62.740,
        62.861,
        62.982,
        66.647,
        70.312,
        68.507,
        66.703,
        68.333,
        69.963,
        85.946,
        101.929,
        106.911,
        111.894,
        112.346,
        112.798,
        107.945,
        103.092,
        112.145,
        121.198,
        127.104,
        133.010,
        132.682,
        132.355,
        129.838,
        127.322,
        127.061,
        126.800,
        122.291,
        117.783,
        117.186,
        116.589,
        115.146,
        113.702,
        111.181,
        108.659,
        109.552,
        110.445,
        108.367,
        106.289,
        105.596,
        104.904,
        102.452,
        100.000,
        97.808,
        95.616,
        94.914,
        94.213,
        90.605,
        86.997,
        87.112,
        87.227,
        86.684,
        86.140,
        84.861,
        83.581,
        81.164,
        78.747,
        78.587,
        78.428,
        76.614,
        74.801,
        74.562,
        74.324,
        74.873,
        75.422,
        73.499,
        71.576,
        67.714,
        63.852,
        64.464,
        65.076,
        66.573,
        68.070,
        62.256,
        56.443,
        60.343,
        64.242,
        66.697,
        69.151,
        63.890,
        58.629,
        50.623,
        42.617,
        51.985,
        61.352,
        59.838,
        58.324,
        0.000,
        0.000,
        0.000,
        0.000,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_d75 = illuminant<illuminant_d75_data>;



struct illuminant_f1_data {
	static constexpr std::size_t min_wl = 380;
	static constexpr std::size_t max_wl = 780;
	static constexpr float normalizer = 1.f / 684.33f;
	static constexpr auto data = std::array<float, 81>{
		1.87f,
		2.36f,
		2.94f,
		3.47f,
		5.17f,
		19.49f,
		6.13f,
		6.24f,
		7.01f,
		7.79f,
		8.56f,
		43.67f,
		16.94f,
		10.72f,
		11.35f,
		11.89f,
		12.37f,
		12.75f,
		13.f,
		13.15f,
		13.23f,
		13.17f,
		13.13f,
		12.85f,
		12.52f,
		12.2f,
		11.83f,
		11.5f,
		11.22f,
		11.05f,
		11.03f,
		11.18f,
		11.53f,
		27.74f,
		17.05f,
		13.55f,
		14.33f,
		15.01f,
		15.52f,
		18.29f,
		19.55f,
		15.48f,
		14.91f,
		14.15f,
		13.22f,
		12.19f,
		11.12f,
		10.03f,
		8.95f,
		7.96f,
		7.02f,
		6.2f,
		5.42f,
		4.73f,
		4.15f,
		3.64f,
		3.2f,
		2.81f,
		2.47f,
		2.18f,
		1.93f,
		1.72f,
		1.67f,
		1.43f,
		1.29f,
		1.19f,
		1.08f,
		0.96f,
		0.88f,
		0.81f,
		0.77f,
		0.75f,
		0.73f,
		0.68f,
		0.69f,
		0.64f,
		0.68f,
		0.69f,
		0.61f,
		0.52f,
		0.43f,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_f1 = illuminant<illuminant_f1_data>;


struct illuminant_f2_data {
	static constexpr std::size_t min_wl = 380;
	static constexpr std::size_t max_wl = 780;
	static constexpr float normalizer = 1.f / 594.5f;
	static constexpr auto data = std::array<float, 81>{
		1.1800f,
   		1.4800f,
   		1.8400f,
   		2.1500f,
   		3.4400f,
  		15.6900f,
   		3.8500f,
   		3.7400f,
   		4.1900f,
   		4.6200f,
   		5.0600f,
  		34.9800f,
  		11.8100f,
   		6.2700f,
   		6.6300f,
   		6.9300f,
   		7.1900f,
   		7.4000f,
   		7.5400f,
   		7.6200f,
   		7.6500f,
   		7.6200f,
   		7.6200f,
   		7.4500f,
   		7.2800f,
   		7.1500f,
   		7.0500f,
   		7.0400f,
   		7.1600f,
   		7.4700f,
   		8.0400f,
   		8.8800f,
  		10.0100f,
  		24.8800f,
  		16.6400f,
  		14.5900f,
  		16.1600f,
  		17.5600f,
  		18.6200f,
  		21.4700f,
  		22.7900f,
  		19.2900f,
  		18.6600f,
  		17.7300f,
  		16.5400f,
  		15.2100f,
  		13.8000f,
  		12.3600f,
  		10.9500f,
   		9.6500f,
   		8.4000f,
   		7.3200f,
   		6.3100f,
   		5.4300f,
   		4.6800f,
   		4.0200f,
   		3.4500f,
   		2.9600f,
   		2.5500f,
   		2.1900f,
   		1.8900f,
   		1.6400f,
   		1.5300f,
   		1.2700f,
   		1.1000f,
   		0.9900f,
   		0.8800f,
   		0.7600f,
   		0.6800f,
   		0.6100f,
   		0.5600f,
   		0.5400f,
   		0.5100f,
   		0.4700f,
   		0.4700f,
   		0.4300f,
   		0.4600f,
   		0.4700f,
   		0.4000f,
   		0.3300f,
   		0.2700f,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_f2 = illuminant<illuminant_f2_data>;


struct illuminant_f3_data {
	static constexpr std::size_t min_wl = 380;
	static constexpr std::size_t max_wl = 780;
	static constexpr float normalizer = 1.f / 556.19f;
	static constexpr auto data = std::array<float, 81>{
		0.82f, 
		1.02f, 
		1.26f, 
		1.44f, 
		2.57f, 
		14.36f,
		2.7f,  
		2.45f, 
		2.73f, 
		3.0f,  
		3.28f, 
		31.85f,
		9.47f, 
		4.02f, 
		4.25f, 
		4.44f, 
		4.59f, 
		4.72f, 
		4.8f,  
		4.86f, 
		4.87f, 
		4.85f, 
		4.88f, 
		4.77f, 
		4.67f, 
		4.62f, 
		4.62f, 
		4.73f, 
		4.99f, 
		5.48f, 
		6.25f, 
		7.34f, 
		8.78f, 
		23.82f,
		16.14f,
		14.59f,
		16.63f,
		18.49f,
		19.95f,
		23.11f,
		24.69f,
		21.41f,
		20.85f,
		19.93f,
		18.67f,
		17.22f,
		15.65f,
		14.04f,
		12.45f,
		10.95f,
		9.51f, 
		8.27f, 
		7.11f, 
		6.09f, 
		5.22f, 
		4.45f, 
		3.8f,  
		3.23f, 
		2.75f, 
		2.33f, 
		1.99f, 
		1.7f,  
		1.55f, 
		1.27f, 
		1.09f, 
		0.96f, 
		0.83f, 
		0.71f, 
		0.62f, 
		0.54f, 
		0.49f, 
		0.46f, 
		0.43f, 
		0.39f, 
		0.39f, 
		0.35f, 
		0.38f, 
		0.39f, 
		0.33f, 
		0.28f, 
		0.21f, 
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_f3 = illuminant<illuminant_f3_data>;


struct illuminant_f9_data {
	static constexpr std::size_t min_wl = 380;
	static constexpr std::size_t max_wl = 800;
	static constexpr float normalizer = 1.f / 743.46;
	static constexpr auto data = std::array<float, 101>{
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.00,
        0.90,
        1.12,
        1.36,
        1.60,
        2.59,
        12.80,
        3.05,
        2.56,
        2.86,
        3.30,
        3.82,
        32.62,
        10.77,
        5.84,
        6.57,
        7.25,
        7.86,
        8.35,
        8.75,
        9.06,
        9.31,
        9.48,
        9.61,
        9.68,
        9.74,
        9.88,
        10.04,
        10.26,
        10.48,
        10.63,
        10.78,
        10.96,
        11.18,
        27.71,
        16.29,
        12.28,
        12.74,
        13.21,
        13.65,
        16.57,
        18.14,
        14.55,
        14.65,
        14.66,
        14.61,
        14.50,
        14.39,
        14.40,
        14.47,
        14.62,
        14.72,
        14.55,
        14.40,
        14.58,
        14.88,
        15.51,
        15.47,
        13.20,
        10.57,
        9.18,
        8.25,
        7.57,
        7.03,
        6.35,
        5.72,
        5.25,
        4.80,
        4.29,
        3.80,
        3.43,
        3.12,
        2.86,
        2.64,
        2.43,
        2.26,
        2.14,
        2.02,
        1.83,
        1.61,
        1.38,
        1.12,
        0.00,
        0.00,
        0.00,
        0.00,
	};

    static constexpr auto coherence_length = 5.f;
	template<typename T>
	static constexpr glm::mat<3,3,T> XYZtoRGB = {
		 2.0413690, -0.5649464, -0.3446944,
		-0.9692660,  1.8760108,  0.0415560,
		 0.0134474, -0.1183897,  1.0154096,
	};
};
using illuminant_f9 = illuminant<illuminant_f9_data>;
